/*
 * Copyright (c) 2006 Pavel Rimsky
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup sparc64	
 * @{
 */
/**
 * @file
 * @brief Routines for writing characters directly to Simics CLI.
 * 
 */

#include <console/chardev.h>
#include <console/console.h>
#include <synch/spinlock.h>
#include <arch/drivers/simics_output.h>

/** maximum number of characters stored before printed */
#define BUFSIZE	512

/* %g2 will store a magic value used during initialization */
asm (".register %g2, #scratch");

/* %g3 will store the address (to be passed to Simics) of the buffer */
asm (".register %g3, #scratch");

/* lock protecting the character buffer */
SPINLOCK_INITIALIZE(simics_buf_lock);

static void simics_putchar(struct chardev * cd, char c);

/** character device operations - only writing will be supported */
static chardev_operations_t simics_stdout_ops = {
	.suspend = NULL,
	.resume = NULL,
	.write = simics_putchar,
	.read = NULL
};

/** Simics character device */
chardev_t simics_stdout;

/**
 * buffer to which output characters are stored before they're printed by Simics
 */
static volatile char buffer[BUFSIZE];

/** Writes a single character to the Simics CLI.
 *
 * The character is not written immediately, but it is stored to the first free
 * position in the buffer, waiting for Simics' Python routine to hang it up
 * and print it. 
 */
static void simics_putchar(struct chardev * cd, char c)
{
	/* the first free position in the buffer */
	static uint16_t current = 0;

	/* '\0' terminates a contiguous block of characters to be printed! */
	if (c == '\0')
		return;

	/* wait till buffer is non-full and other processors aren't writing to it */
	while (1) {
		while (buffer[current] != 0)
			;
		if (spinlock_trylock(&simics_buf_lock))
			break;
	}

	buffer[current] = c;
	
	current = (current + 1) % BUFSIZE;
	membar();
	
	spinlock_unlock(&simics_buf_lock);
}

/** Initializes the Simics output.
 * 
 * Passes the address of the buffer to the Simics' Python script and
 * redirects kernel output to the Simics CLI. 
 */
void simics_output_init(void)
{
	/* all buffer positions are free at the beginning */
	uint16_t i;
	for (i = 0; i < BUFSIZE; i++) {
		buffer[i] = '\0';
	}

	/*
	 * pass the address of the buffer to the Simics' Python script
	 *   - write it to the %g3 register
	 *   - write the magic value to the %g2 register
	 *     (so that the script knows that the value in %g3 is valid)
	 *   - loop until the value is read
	 *     (the script notifies us by setting %g2 to 0) 
	 */
	asm volatile (
		"or %0, 0, %%g3\n"
		
		"set 0x18273645, %%g2\n"
		
		"0: cmp %%g2, 0\n"
		"bnz 0b\n"
		"nop"
		:: "r" (buffer)
	);
				
	/* redirect kernel output */
	chardev_initialize("simics_output", &simics_stdout, &simics_stdout_ops);
	stdout = &simics_stdout;
}

/** @}
 */
